"""Full feature class definition."""
import datetime
import typing


class Owner:
    """Class that will be referenced in Example"""
    name: str

    def __init__(self, name):
        self.name = name


class Tag:
    """Class that will be referenced in Example"""
    name: str

    def __init__(self, name):
        self.name = name


class Example:
    """This class has annotations with all supported fields."""
    name: str
    amount_round: int
    amount: float
    useful: bool
    created: datetime.datetime
    expiration: datetime.date
    refresh_at: datetime.time
    owner: Owner
    tags: typing.List[Tag]
    redundant: str

    @classmethod
    def create_instance(cls) -> 'Example':
        """Return object instance with some data."""
        instance = cls()
        instance.name = 'Full feature'
        instance.amount = 5.3
        instance.amount_round = 5
        instance.useful = True
        instance.created = datetime.datetime(2019, 2, 6, 15, 0)
        instance.expiration = datetime.date(2050, 12, 31)
        instance.refresh_at = datetime.time(12, 30)
        instance.owner = Owner('me')
        instance.tags = []
        instance.tags.append(Tag('graphene'))
        instance.tags.append(Tag('type annotations'))
        instance.redundant = 'Full feature (again)'
        return instance



"""Nested classes definition."""
from typing import List


class SubLeaf:
    """Even smaller leaf, which is a part of bigger one. Pretty weird, huh?"""
    value: str


class Leaf:
    """Leaf class definition."""
    value: str
    leaflets: List[SubLeaf]


class TopLevel:
    """Top level class which references to Leaf."""
    leaf: Leaf
    name: str



"""Main module."""
import typing
from datetime import date, datetime, time

import graphene

BASIC_TYPE_MAPPINGS = {
    str: graphene.String,
    int: graphene.Int,
    bool: graphene.Boolean,
    float: graphene.Float,
    date: graphene.Date,
    datetime: graphene.DateTime,
    time: graphene.Time,
}

UNSUPORTED_TYPES = (list, dict, tuple, set)


class UnsupportedAnnotationError(Exception):
    """Raised on unsupported annotation resolve attempt."""


class NoAnnotationsError(Exception):
    """Raised when no annotations have been found (or all are excluded)."""


class ObjectType(graphene.ObjectType):
    """
    Base class for type annotated graphene schemas.
    The subclass still has to be decorated, the purpose of this class is to provide hints
    for special graphanno attributes and those inherited from ObjectType.
    """
    __model__: typing.Any = None
    __excluded_fields__: typing.Iterable = tuple()
    __ignore_unsupported__: bool = False


def graph_annotations(cls):
    """Prepare GraphQL schema based on the type annotations."""
    attributes = {}
    target_class = cls.__model__ if hasattr(cls, '__model__') else cls
    ignore_unsupported = getattr(cls, '__ignore_unsupported__', False)
    excluded_keys = getattr(cls, '__excluded_fields__', tuple())

    annotations = dict(**getattr(target_class, '__annotations__', {}))
    annotations.update(getattr(cls, '__annotations__', {}))

    private_keys = tuple(key for key in annotations.keys() if key.startswith('_'))

    for key in excluded_keys + private_keys:
        annotations.pop(key)

    if not annotations:
        raise NoAnnotationsError(f'No included annotations for class {cls.__name__}.')

    for name, annotation in annotations.items():
        if annotation in UNSUPORTED_TYPES:
            if ignore_unsupported:
                continue
            raise UnsupportedAnnotationError(f'The type annotation: {annotation} is not supported.')
        type_, args = _get_type_from_annotation(annotation)
        attributes[name] = type_(*args)

    superclasses = (cls,) if issubclass(cls, graphene.ObjectType) else (cls, graphene.ObjectType)
    return type(cls.__name__, superclasses, attributes)


def _get_type_from_annotation(annotation, type_only=False):
    basic_type = BASIC_TYPE_MAPPINGS.get(annotation)
    if basic_type:
        return basic_type if type_only else (basic_type, tuple())

    if str(annotation).startswith('typing.List'):
        return graphene.List, (_get_type_from_annotation(annotation.__args__[0], True),)

    type_ = graph_annotations(annotation)
    return type_ if type_only else (graphene.Field, (type_,))



"""A setuptools based setup module."""

from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = "".join([line for line in f.readlines() if '.svg' not in line])

setup(
    name='graphanno',
    version='1.1.0',
    description='Create graphene ObjectType based on the type annotations.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/aklajnert/graphanno',
    author='Andrzej Klajnert',
    author_email='py@aklajnert.pl',
    license="MIT",
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Code Generators',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],

    keywords='graphene graphql type_annotations annotations',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    python_requires='>=3.6, <4',
    install_requires=['graphene>=2.0'],
    extras_require={'test': ['pytest', 'pylint', 'flake8'], },
    project_urls={
        'Bug Reports': 'https://github.com/aklajnert/graphanno/issues',
        'Source': 'https://github.com/aklajnert/graphanno',
    },
)
